import os
import re

import yaml
from pydash import _


class ShragaConfig:
    path_matcher = re.compile(r"\$\{([^}^{]+)}")

    def __init__(self):
        self.all_configs = None

    def path_constructor(self, loader, node):
        """Extract the matched value, expand env variable, and replace the match"""
        value = node.value
        match = self.path_matcher.match(value)
        env_var = match.group()[2:-1]
        env_var_val = os.environ.get(env_var)
        if env_var_val:
            return env_var_val + value[match.end() :]
        return ""

    def load(self, config_path: str = os.getenv("CONFIG_PATH") or "config.yaml"):
        print(f"Loading config from {config_path}")
        with open(config_path) as stream:
            yaml.add_implicit_resolver("!path", self.path_matcher)
            yaml.add_constructor("!path", self.path_constructor)
            self.all_configs = yaml.load(stream, Loader=yaml.FullLoader)
            return self

    def get(self, k: str, default=None):
        v = _.get(self.all_configs, k)
        if v is None:
            return default
        return v

    def set(self, k: str, v):
        _.set(self.all_configs, k, v)

    def auth_realms(self):
        return _.get(self.all_configs, "auth.realms") or dict()

    def auth_users(self):
        return set(_.get(self.all_configs, "auth.users") or [])

    def retrievers(self) -> dict:
        return _.get(self.all_configs, "retrievers") or dict()



from .base_search_retriever import BaseSearchRetriever
from .collapse import get_inner_hits, merge_inner_hits_results
from .common import RetrieverConfig
from .elasticsearch import ElasticsearchRetriever
from .get_client import get_client

# commented out due to slow loading!
# from .google_maps import GoogleMapsRetriever
from .local import LocalRetriever
from .opensearch import OpenSearchRetriever

__all__ = [
    "ElasticsearchRetriever",
    "OpenSearchRetriever",
    "LocalRetriever",
    "BaseSearchRetriever",
    "RetrieverConfig",
    "merge_inner_hits_results",
    "get_inner_hits",
    "get_client",
]



class RequestCancelledException(Exception):
    pass

class LLMServiceUnavailableException(Exception):
    def __init__(self, msg, orig_exception = None):
        extra_info = ""
        if orig_exception:
            exception_type = type(orig_exception).__name__
            exception_message = str(orig_exception)
            extra_info = ",".join([m for m in [exception_type, exception_message] if m])
            extra_info = f"({extra_info})"
        full_message = f"{msg} {extra_info}".strip()
        super().__init__(full_message)



from shraga_common import ShragaConfig
from shraga_common.retrievers import get_client


def get_history_client(shraga_config: ShragaConfig):
    history_enabled = shraga_config.get("history.enabled", True)
    if history_enabled:
        client = get_client(shraga_config)
        index = shraga_config.get("history.index") or "chat-history"
    else:
        client = None
        index = None

    return client, index



import asyncio
import json
from abc import abstractmethod
from datetime import datetime
from typing import List, Optional

from shraga_common.app import RequestCancelledException, LLMServiceUnavailableException
from shraga_common.models import (FlowBase, FlowResponse, FlowRunRequest,
                                  HistoryMessage)
from shraga_common.services import (BedrockService, LLMService,
                                    LLMServiceOptions, OpenAIService)
from shraga_common.services.llm_service import LlmModelProvider


class LLMFlowBase(FlowBase):
    llmservice: LLMService = None
    listed = False
    llm_model_provider: LlmModelProvider = None
    llm_model_name: str = None
    parse_json: bool = True

    @abstractmethod
    def get_prompt(self):
        pass

    @staticmethod
    def format_chat_history(history: List[HistoryMessage], history_window: int):
        if not history or history_window == 0:
            return ""

        history_txt = []
        for msg in history:
            text = msg.text.strip()
            if msg.msg_type == "system":
                history_txt.append(f"Assistant: {text}\n\n")
            elif msg.msg_type == "user":
                history_txt.append(f"User: {text}\n\n")

        return "".join(history_txt[-history_window:])

    def init_model(self):
        provider = self.llm_model_provider
        if not self.llmservice:
            if provider == "cohere":
                self.llmservice = BedrockService(self.config)
            elif provider == "openai":
                self.llmservice = OpenAIService(self.config)

    def format_prompt(self, question: str, format_info: Optional[dict] = None):
        prompt = self.get_prompt()
        format_info = format_info or {}
        question = (
            format_info.get("question", question)
            .replace('"', '\\"')
            .replace("'", "\\'")
        )

        today = format_info.get("last_lookup_date", datetime.now())
        current_year = today.year
        current_date = today.strftime("%Y-%m-%d")

        format_info.update(
            {
                "question": question,
                "current_date": current_date,
                "current_year": current_year,
            }
        )
        return prompt.format(
            **format_info,
        )

    async def run_prompt(self, prompt: str):
        self.init_model()
        content = None
        response_text = ""
        payload = {}
        try:
            start_time = datetime.now()
            llm_context: LLMServiceOptions = {}
            if self.llm_model_name:
                llm_context["model_id"] = self.llm_model_name

            try:
                content = await self.llmservice.invoke_model(prompt, llm_context)
                run_time = datetime.now() - start_time
                self.trace(f"execute runtime: {run_time}")
                if self.parse_json:
                    payload = json.loads(content.text, strict=False)
                else:
                    response_text = content.text

            except asyncio.CancelledError:
                raise RequestCancelledException("LLM flow cancelled")
            except LLMServiceUnavailableException:
                raise

        except (RequestCancelledException, LLMServiceUnavailableException):
            raise
        except Exception as e:
            payload = {"error": str(e), "body": content if content else ""}

        return response_text, payload, content

    async def execute(self, request: FlowRunRequest) -> FlowResponse:
        self.trace(f"using {self.llm_model_provider}\{self.llm_model_name}")
        prompt = self.format_prompt(
            request.question,
            {
                **(
                    request.context
                    if isinstance(request.context, dict)
                    else request.context.dict()
                ),
                "chat_history": request.chat_history,
                "history_window": request.preferences.get("history_window", 0),
            },
        )
        response_text, payload, content = await self.run_prompt(prompt)

        return FlowResponse(
            response_text=response_text,
            payload=payload,
            trace=self.trace_log,
            stats=[self.get_stats(content)],
        )

    @staticmethod
    def id():
        return "llm-routing-v1"

    @staticmethod
    def description():
        return "default llm powered routing flow."



import asyncio
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from .common import RetrieverConfig


class BaseSearchRetriever(ABC):
    def __init__(self, *args, **kwargs):
        self.client = None

    @staticmethod
    @abstractmethod
    def get_client(shraga_config, extra_configs: RetrieverConfig):
        pass

    @abstractmethod
    async def get_indices_list(self):
        pass

    @abstractmethod
    async def execute_vector_search(
        self,
        field_name: str,
        query_vector: List[float],
        k: int = 10,
        index_name: Optional[str] = None,
    ) -> List[Dict]:
        pass

    @abstractmethod
    async def execute_text_search(
        self, text: str, k: int = 10, index_name: Optional[str] = None
    ) -> List[Dict]:
        pass

    async def execute_empty_query(self, index_name: Optional[str] = None):
        pass

    def count(self, body: dict, index_name: str):
        return self.client.count(body, index_name)

    @abstractmethod
    def execute_with_timeout(self, body: dict, index_name: str, timeout):
        pass

    @abstractmethod
    async def execute_raw_search(
        self, raw_query: dict, index_name: Optional[str] = None
    ) -> List[Dict]:
        pass

    async def execute_raw_searches(
        self, queries: List[dict], index_name: Optional[str] = None
    ) -> List[Dict]:
        tasks = [self.execute_raw_search(query, index_name) for query in queries]
        return await asyncio.gather(*tasks)



import json
import os
from collections import namedtuple
from datetime import datetime
from typing import List, Optional, Union

import boto3
import geopandas as gpd
import googlemaps
import pandas as pd
import requests
from botocore.exceptions import ClientError
from shapely.geometry import Point, Polygon
from tqdm.auto import tqdm

from shraga_common import ShragaConfig


class GoogleMapsRetriever:
    gmaps = None
    Place = None
    gdf_data = None

    def __init__(
        self,
        shraga_config: ShragaConfig,
        fn_cities_and_urban_areas: str,
        retriever_name: Optional[str] = None,
    ):
        self.fn_cities_and_urban_areas = fn_cities_and_urban_areas
        self.config = shraga_config
        self.api_key = shraga_config.get("services.googlemaps.api_key")
        self.gmaps = googlemaps.Client(key=self.api_key)
        self.Place = namedtuple("Place", ["name", "type", "lat", "lon"])
        self.initGdf()

    def initGdf(self):
        aws_region = self.config.get("aws.region") or "us-east-1"
        aws_bucket = self.config.get("aws.assets_bucket")
        if not os.path.exists(self.fn_cities_and_urban_areas) and aws_bucket:
            try:
                s3 = boto3.resource("s3", region_name=aws_region)
                s3.Bucket(aws_bucket).download_file(
                    "cities_and_urban_areas.geojson", self.fn_cities_and_urban_areas
                )
            except ClientError as e:
                raise FileNotFoundError(
                    f"File not found: {self.fn_cities_and_urban_areas}. Asset download failed with "
                    + str(e.response["Error"]["Code"])
                )
            except Exception as e:
                raise FileNotFoundError(
                    f"File not found: {self.fn_cities_and_urban_areas}. Asset download failed with "
                    + str(e)
                )

        if not os.path.exists(self.fn_cities_and_urban_areas):
            raise FileNotFoundError(f"File not found: {self.fn_cities_and_urban_areas}")

        self.gdf_data = gpd.read_file(self.fn_cities_and_urban_areas)

    # Google Maps API utilities
    def get_location_by_name(self, name: str) -> Optional[dict]:
        "Given a location name, return the geocoded location."

        geocode_result = self.gmaps.geocode(name)
        return geocode_result[0] if geocode_result else None

    def get_fastest_route(self, origin, destination):
        """
        Get the fastest route between two points
        """
        # Get the current time for real-time traffic estimation
        now = datetime.now()

        # Get directions with real-time traffic consideration
        directions_result = self.gmaps.directions(
            origin=origin,
            destination=destination,
            mode="driving",
            departure_time=now,
            traffic_model="best_guess",  # Estimate based on real-time traffic conditions
        )

        # Check if a route was found
        if not directions_result:
            return None

        # Extract and display relevant information
        for route in directions_result:
            leg = route["legs"][0]
            route_info = {
                "start": leg["start_address"],
                "end": leg["end_address"],
                "distance": leg["distance"]["text"],
                "duration_without_traffic": leg["duration"]["text"],
                "duration_with_traffic": leg["duration_in_traffic"]["text"],
                "steps": [step["html_instructions"] for step in leg["steps"]],
            }
            return route_info

    def get_routes_between_locations(
        self, loc_a: str, loc_b: str, n_routes: int
    ) -> List[dict]:
        "Given two locations (as strings), return the routes between them."
        directions_result = self.gmaps.directions(loc_a, loc_b, alternatives=True, units="metric")
        return directions_result[:n_routes]

    def get_locations_along_route(
        self,
        loc_a: str,
        loc_b: str,
        n_routes: int = 2,
        radius_km: float = 2.0,
        location_types: Union[None, list, set, str] = "selected",
        step_fraction: float = 1.0,
    ) -> List[Place]:
        """
        Given two locations, return the places along the routes between them.

        The function first retrieves the routes between the two locations using the `get_routes_between_locations` function.
        It then iterates over the routes and extracts the steps along each route. For each step, the function retrieves the
        places nearby using the Google Places API. The function filters the places based on the specified location types and
        radius. The function returns a list of named tuples representing the places along the routes.

        Args:
            loc_a (str): The starting location.
            loc_b (str): The destination location.
            n_routes (int): The number of alternative routes to retrieve.
            radius_km (float, optional): The radius in kilometers to search for places around each point. Defaults to 2.0.
            location_types (Union[None, list, set, str], optional): Types of locations to search for.
                Use "selected" for default types,
                None for all types, or provide a list/set of specific types. Defaults to "selected".
            step_fraction (float, optional): Fraction of steps to consider along the route (0.0 to 1.0). Defaults to 1.0.

        Returns:
            List[Place]: A list of named tuples representing the places along the routes.


        """
        if location_types == "selected":
            location_types = {"locality"}  # , "route", "intersection"}
        elif isinstance(location_types, str):
            location_types = location_types.split(",")
        if location_types is None:
            location_types = [None]

        routes = self.get_routes_between_locations(loc_a, loc_b, n_routes)
        places_along_routes = []

        for route in tqdm(routes, desc="Processing routes"):
            path = []
            for leg in tqdm(route["legs"], desc="Processing route legs"):
                steps = leg["steps"]
                num_steps = max(1, int(len(steps) * step_fraction))
                for step in steps[:num_steps]:
                    lat_lng = step["end_location"]
                    path.append((lat_lng["lat"], lat_lng["lng"]))

            for point in tqdm(path, desc="Processing points along route"):
                for location_type in location_types:
                    places_results = self.gmaps.places_nearby(
                        location=point,
                        radius=radius_km * 1_000,
                        type=location_type,
                        language="en",
                    )
                    if places_results["results"]:
                        for pr in places_results["results"]:
                            curr = self.Place(
                                name=pr["name"],
                                type=",".join(pr["types"]),
                                lat=pr["geometry"]["location"]["lat"],
                                lon=pr["geometry"]["location"]["lng"],
                            )
                            if curr not in places_along_routes:
                                places_along_routes.append(curr)
        return places_along_routes

    def get_places_near_location(
        self,
        location: Union[str, dict],
        radius_km: float = 2.0,
        location_types: Union[None, list, set, str] = "selected",
        limit=20,
    ) -> List[Place]:
        if radius_km > 50:
            raise ValueError(
                "The maximum radius allowed by the Google Places API is 50 km."
            )
        if limit > 20:
            raise ValueError(
                "The maximum number of results allowed by the Google Places API is 20."
            )

        if location_types == "selected":
            location_types = ["locality"]  # , "route", "intersection"]
        elif isinstance(location_types, str):
            location_types = location_types.split(",")
        if location_types is None:
            location_types = [None]

        if isinstance(location, str):
            loc = self.get_location_by_name(location)
            if not loc:
                return []
            lat, lon = (
                loc["geometry"]["location"]["lat"],
                loc["geometry"]["location"]["lng"],
            )
        elif isinstance(location, dict):
            lat, lon = location["lat"], location["lon"]
        else:
            raise ValueError(
                "The location should be a string or a dictionary with 'lat' and 'lon' keys."
            )

        places = []

        for location_type in location_types:
            endpoint = "https://places.googleapis.com/v1/places:searchNearby"
            body = {
                "includedTypes": [location_type],
                "maxResultCount": limit,
                "locationRestriction": {
                    "circle": {
                        "center": {"latitude": lat, "longitude": lon},
                        "radius": radius_km * 1000,
                    }
                },
            }
            headers = {
                "Content-Type": "application/json",
                "X-Goog-Api-Key": self.api_key,
                "X-Goog-FieldMask": "places.displayName,places.location,places.types",
            }
            response = requests.post(endpoint, headers=headers, data=json.dumps(body))
            places_results = response.json()

            if "places" in places_results:
                for pr in places_results["places"]:
                    curr = self.Place(
                        name=pr["displayName"]["text"],
                        type=",".join(pr["types"]),
                        lat=pr["location"]["latitude"],
                        lon=pr["location"]["longitude"],
                    )
                    if curr not in places:
                        places.append(curr)
        return places

    def generate_locations_from_a_country(
        self,
        country_code,
        limit=10,
    ) -> list:
        assert (
            len(country_code) == 2
        ), f"Country code must be a two-letter code. You provided: {country_code}"
        country_code = country_code.upper()

        gdf = pd.DataFrame(
            self.gdf_data[self.gdf_data["country code"] == country_code]
            .sort_values("population", ascending=False)
            .head(limit)
        )

        places = []
        for _, row in gdf.iterrows():
            geometry = row["geometry"] if "geometry" in row else ""
            if geometry.geom_type == "Point":
                centroid = (geometry.y, geometry.x)
            elif geometry.geom_type == "Polygon":
                centroid = gpd.GeoSeries(geometry).centroid
            else:
                raise Exception(f"Unknown geometry type {geometry.geom_type}")
            places.append(
                self.Place(
                    name=row["name"],
                    type="locality",
                    lat=centroid[0],
                    lon=centroid[1],
                )
            )

        return places

    def is_within_distance_of_point(
        self, polygons_gdf: gpd.GeoSeries, point_lon_lat: Point, distance_km: float
    ):
        distance_m = distance_km * 1000
        point_gdf = gpd.GeoDataFrame(
            index=[0], crs="EPSG:4326", geometry=[point_lon_lat]
        )
        polygons_gdf = polygons_gdf.set_crs("EPSG:4326").to_crs(epsg=3857)
        point_gdf = point_gdf.to_crs(epsg=3857)
        point_buffer = point_gdf.buffer(distance_m).iloc[0]
        return polygons_gdf.intersects(point_buffer)

    def is_within_distance_of_polygon(
        self, polygons_gdf: gpd.GeoSeries, polygon_lon_lat: Polygon, distance_km: float
    ):
        distance_m = distance_km * 1000
        polygon_gdf = gpd.GeoDataFrame(
            index=[0], crs="EPSG:4326", geometry=[polygon_lon_lat]
        )
        polygons_gdf = polygons_gdf.set_crs("EPSG:4326").to_crs(epsg=3857)
        polygon_gdf = polygon_gdf.to_crs(epsg=3857)
        polygon_buffer = polygon_gdf.buffer(distance_m).iloc[0]
        return polygons_gdf.intersects(polygon_buffer)

    def distance_between(
        self,
        a_lon_lat: Union[Point, Polygon, tuple],
        b_lon_lat: Union[Point, Polygon, tuple],
    ) -> float:
        if isinstance(a_lon_lat, tuple):
            a_lon_lat = Point(a_lon_lat[1], a_lon_lat[0])
        if isinstance(b_lon_lat, tuple):
            b_lon_lat = Point(b_lon_lat[1], b_lon_lat[0])

        gdf_a = gpd.GeoDataFrame(
            index=[0], crs="EPSG:4326", geometry=[a_lon_lat]
        ).to_crs(epsg=3857)
        gdf_b = gpd.GeoDataFrame(
            index=[0], crs="EPSG:4326", geometry=[b_lon_lat]
        ).to_crs(epsg=3857)
        distance_m = gdf_a.distance(gdf_b).iloc[0]
        return distance_m / 1000

    def find_interesting_locations(
        self,
        city_a: str,
        city_b: Optional[str] = None,
        n_routes: int = 2,
        radius_km: float = 10.0,
        location_types: str = "selected",
        step_fraction: float = 1.0,
        output_file: Optional[str] = None,
        verbose: bool = True,
    ) -> None:
        if city_b:
            result = self.get_locations_along_route(
                city_a, city_b, n_routes, radius_km, location_types, step_fraction
            )
        else:
            result = self.get_places_near_location(city_a, radius_km, location_types)
        result = pd.DataFrame(result)
        if verbose:
            print(f"The places found are:\n{result.to_markdown()}")
        if output_file:
            result.to_csv(output_file, index=False)
            if verbose:
                print(f"Results saved to {output_file}")

    def demonstrate(
        self,
        *,
        city_a: str,
        city_b: str = None,
        n_routes: int = 2,
        radius_km: float = 10.0,
        location_types: str = "selected",
        step_fraction: float = 1.0,
        output_file: str = None,
        verbose: bool = False,
    ) -> None:
        """
        Find interesting locations along routes or around a city using Google Maps API.

        Args:
            city_a (str): Name of the first city (required).
            city_b (str, optional): Name of the second city. Defaults to None.
            n_routes (int, optional): Number of alternative routes to consider. Defaults to 2.
            radius_km (float, optional): Radius in kilometers to search for places. Defaults to 10.0.
            location_types (str, optional): Types of locations to search for. Defaults to "selected".
            step_fraction (float, optional): Fraction of steps to consider along the route. Defaults to 1.0.
            output_file (str, optional): Path to save the output to a file. Defaults to None.
            verbose (bool, optional): Print verbose output. Defaults to False.
        """
        self.find_interesting_locations(
            city_a=city_a,
            city_b=city_b,
            n_routes=n_routes,
            radius_km=radius_km,
            location_types=location_types,
            step_fraction=step_fraction,
            output_file=output_file,
            verbose=verbose,
        )



from logging import getLogger
from typing import Dict, List, Optional

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError
from pydash import _

from shraga_common import ShragaConfig

from .base_search_retriever import BaseSearchRetriever
from .common import RetrieverConfig

logger = getLogger(__name__)


class ElasticsearchRetriever(BaseSearchRetriever):
    """
    ElasticsearchRetriever class
    """

    def __init__(self, shraga_config: ShragaConfig):
        super().__init__()
        config: RetrieverConfig = shraga_config.get("retrievers.elasticsearch")
        self.client = ElasticsearchRetriever.get_client(shraga_config, config)
        self.index_name = config.get("index")

    @staticmethod
    def get_client(shraga_config, extra_configs: RetrieverConfig):
        use_cloud_id = extra_configs.get("use_cloud_id", "").lower() == "true"
        cloud_id = extra_configs.get("cloud_id")
        host = extra_configs.get("host")
        port = extra_configs.get("port", 9200)
        use_ssl = (
            False
            if extra_configs.get("use_ssl", False) in {False, "false", "False"}
            else True
        )
        verify_certs = extra_configs.get("verify_certs", False)

        # Authentication method
        auth_type = extra_configs.get("auth_type", "basic").lower()
        if auth_type == "apikey":
            api_key = extra_configs.get("api_key")
            auth = {"api_key": api_key} if api_key else None
        else:  # default to basic auth
            http_auth_user = extra_configs.get("user")
            http_auth_password = extra_configs.get("password")
            auth = (
                {"basic_auth": (http_auth_user, http_auth_password)}
                if http_auth_user and http_auth_password
                else None
            )

        if use_cloud_id:
            if not cloud_id:
                raise ValueError("cloud_id must be provided if use_cloud_id is true")
            return Elasticsearch(
                cloud_id=cloud_id, use_ssl=use_ssl, verify_certs=verify_certs, **auth
            )
        else:
            if not host:
                raise ValueError("host must be provided if use_cloud_id is false")
            return Elasticsearch(
                hosts=[f"{'https' if use_ssl else 'http'}://{host}:{port}"],
                # use_ssl=use_ssl,
                verify_certs=verify_certs,
                **auth,
            )

    async def get_indices_list(self):
        pass

    def get_doc_by_id(self, document_id: str, index_name: Optional[str] = None) -> Dict:
        try:
            # Retrieve the document
            response = self.client.get(index=index_name, id=document_id)
            return response["_source"]  # Return the document's source data
        except NotFoundError:
            print(f"Document with ID {document_id} not found in index {index_name}.")
            return None
        except Exception as e:
            print(f"An error occurred: {e}")
            return None

    async def execute_empty_query(self, index_name: Optional[str] = None):
        return await self.execute_raw_search(
            {"query": {"match_all": {}}, "size": 0}, index_name
        )

    async def execute_vector_search(
        self,
        field_name: str,
        query_vector: List[float],
        k: int = 10,
        index_name: Optional[str] = None,
    ) -> List[Dict]:
        query = {
            "size": k,  # max number of results to return
            "query": {
                "knn": {
                    field_name: {
                        "vector": query_vector,
                        "k": k,  # number of nearest neighbors to return
                    }
                }
            },
        }
        return await self.execute_raw_search(query, index_name)

    async def execute_text_search(
        self,
        text: str,
        k: int = 10,
        index_name: Optional[str] = None,
    ) -> List[Dict]:
        query = {"size": k, "query": {"match": {"content": text}}}
        return await self.execute_raw_search(query)

    async def execute_raw_search(
        self, raw_query: dict, index_name: Optional[str] = None
    ) -> List[Dict]:
        try:
            response = self.client.search(
                index=index_name or self.index_name, body=raw_query
            )
            # TODO validate search response
            hits = _.get(response, "hits.hits") or []
            return hits
        except Exception as e:
            logger.error(f"Error executing search query {e}")
            raise e

    def execute_with_timeout(self, body: dict, index_name: str, timeout):
        pass



import logging
from abc import abstractmethod
from typing import List, Literal, Optional, get_args

from ..logging import init_logging
from ..shraga_config import ShragaConfig
from .flow_response import FlowResponse
from .flow_run_request import FlowRunRequest
from .retrieval_result import RetrievalResult

logger = init_logging(__name__)


class FlowBase:
    listed = True

    def __init__(self, config: ShragaConfig, flows: Optional[dict] = None):
        self.config = config
        self.flows = flows
        self.trace_log = dict()
        self.trace_log["log"] = []

        if self.config.get("debug.print"):
            logger.setLevel(logging.DEBUG)

    @abstractmethod
    async def execute(self, request: FlowRunRequest) -> FlowResponse:
        raise NotImplementedError()

    @staticmethod
    def id() -> str:
        raise NotImplementedError()

    @staticmethod
    def description() -> str:
        return ""

    @staticmethod
    def format_pydantic_config(model):
        config = {}

        for field_name, field in model.model_fields.items():
            field_info = {
                "default_value": field.default,
            }

            if field.description:
                field_info["description"] = field.description

            field_type = field.annotation
            if hasattr(field_type, "__origin__") and field_type.__origin__ is Literal:
                allowed_values = get_args(field_type)
                field_info["available_values"] = list(allowed_values)
                field_info["type"] = (
                    "string"  # assuming Literal fields are string-based
                )
            else:
                if field_type is int:
                    field_info["type"] = "integer"
                elif field_type is str:
                    field_info["type"] = "string"
                elif field_type is float:
                    field_info["type"] = "float"
                elif field_type is bool:
                    field_info["type"] = "boolean"

            config[field_name] = field_info

        return config

    @staticmethod
    def available_preferences(config) -> dict:
        return dict()

    @staticmethod
    def get_tool_desc():
        return None

    @staticmethod
    def get_flow_instance(flows: list, flow_id: str, config: ShragaConfig):
        f = flows.get(flow_id)
        if not f:
            raise RuntimeError(f"No flow found with id {flow_id}")
        return f(config, flows)

    async def execute_another_flow_by_id(
        self, flow_id: str, request: FlowRunRequest
    ) -> FlowResponse:
        f = self.flows.get(flow_id)
        if not f:
            raise RuntimeError(f"No flow found with id {flow_id}")

        flow_instance = f(self.config, self.flows)
        response = await flow_instance.execute(request)
        if response.trace and "log" in response.trace:
            self.trace_log["log"].extend(response.trace["log"])
        return response

    async def execute_another_flow(
        self, flow_clazz, request: FlowRunRequest
    ) -> FlowResponse:
        f = self.flows.get(flow_clazz().id())
        if not f:
            raise RuntimeError("No flow found")
        return await f(self.config, self.flows).execute(request)

    def trace(self, text: str = ""):
        flow_id = type(self).id()
        text = f"{flow_id}: {text}"
        log = self.trace_log.get("log") or []
        log.append(text)
        self.trace_log["log"] = log
        if self.config.get("debug.print"):
            logger.debug(text)

    def get_stats(self, content):
        if not content:
            return {}
        
        if isinstance(content, dict):
            stats = content
        else: 
            stats = content.stats.dict()
        stats["flow_id"] = type(self).id()
        return stats

    def response(
        self, text: str, results: List[RetrievalResult], payload: Optional[dict] = None
    ) -> FlowResponse:
        return FlowResponse(
            response_text=text,
            retrieval_results=results,
            payload=payload or dict(),
            trace=self.trace_log,
        )



import unittest
from unittest.mock import AsyncMock

from shraga_common import ShragaConfig
from shraga_common.services import BedrockService

from .evaluation import EvaluationFlow, EvaluationModel

testcase = [
    {
        "testcase": {
            "question": "What is the capital of France?",
            "answer": "Paris",
        },
        "evaluation": {"run_time": 0.1},
        "generated_answer": "Paris",
    }
]
preferences = EvaluationModel(
    input_files=[],
    flow_id="test",
)


class FlowResponseMock:
    text: str

    def __init__(self, text: str):
        self.text = text


class EvaluationFlowDocHandler(unittest.IsolatedAsyncioTestCase):
    """
    Test suite for MaxDocHandler class.
    Tests various document types and their processing logic.
    """

    async def asyncSetUp(self):
        """Initialize test configuration and handler"""
        self.shraga_config = ShragaConfig()
        self.flow = EvaluationFlow(self.shraga_config)
        self.flow.llmservice = BedrockService(self.shraga_config)

    async def test_empty_eval_results(self):
        res = await self.flow.evaluate_results([], preferences)
        self.assertIsNone(res)

    async def test_correct_eval_results(self):
        self.flow.llmservice.invoke_model = AsyncMock(
            return_value=FlowResponseMock(text='{"correctness": "correct"}')
        )
        res = await self.flow.evaluate_results(testcase, preferences)
        self.assertTrue(res["stats"]["total_scenarios"] == 1)
        self.assertTrue(res["stats"]["correct_count"] == 1)
        self.assertTrue(res["stats"]["correct_percentage"] == 100)
        self.assertTrue(len(res["results"]) == 1)
        self.assertTrue(res["results"][0]["evaluation"]["correctness"] == "correct")

    async def test_partial_eval_results(self):
        self.flow.llmservice.invoke_model = AsyncMock(
            return_value=FlowResponseMock(text='{"correctness": "partially correct"}')
        )
        res = await self.flow.evaluate_results(testcase, preferences)
        self.assertTrue(res["stats"]["total_scenarios"] == 1)
        self.assertTrue(res["stats"]["partial_correct_count"] == 1)
        self.assertTrue(res["stats"]["correct_percentage"] == 0)
        self.assertTrue(res["stats"]["partial_correct_percentage"] == 100)
        self.assertTrue(len(res["results"]) == 1)

    async def test_incorrect_eval_results(self):
        self.flow.llmservice.invoke_model = AsyncMock(
            return_value=FlowResponseMock(text='{"correctness": "incorrect"}')
        )
        res = await self.flow.evaluate_results(testcase, preferences)
        self.assertTrue(res["stats"]["total_scenarios"] == 1)
        self.assertTrue(res["stats"]["correct_count"] == 0)
        self.assertTrue(res["stats"]["correct_percentage"] == 0)
        self.assertTrue(len(res["results"]) == 1)

    async def test_corrupt_eval_results(self):
        self.flow.llmservice.invoke_model = AsyncMock(
            return_value=FlowResponseMock(text='{"correctness": "}')
        )
        res = await self.flow.evaluate_results(testcase, preferences)
        self.assertTrue(res["stats"]["total_scenarios"] == 1)
        self.assertTrue(res["stats"]["correct_count"] == 0)
        self.assertTrue(res["stats"]["correct_percentage"] == 0)
        self.assertTrue(len(res["results"]) == 1)
        self.assertIsNone(res["results"][0]["evaluation"].get("correctness"))



import asyncio

from shraga_common.app import RequestCancelledException
from shraga_common.models import (FlowBase, FlowResponse, FlowRunRequest,
                                  RetrievalResult)

"""
A demo flow class that simulates response delay and returns mock retrieval results.
Mainly used for testing and demonstration purposes.
"""


class DemoDefaultFlow(FlowBase):

    async def execute(self, request: FlowRunRequest) -> FlowResponse:

        try:
            answer = f'You asked "{request.question}"'

            total_delay = 3
            check_interval = 0.5

            for i in range(int(total_delay / check_interval)):
                await asyncio.sleep(check_interval)

            formatted_results = [
                RetrievalResult(
                    id="1",
                    title=f"Result for question: {request.question}",
                    description=f"Description for question: {request.question}",
                    score=0.9,
                )
            ]

            return self.response(
                text=answer, results=formatted_results, payload={"answer": answer}
            )

        except asyncio.CancelledError:
            raise RequestCancelledException("Flow cancelled")
        except Exception:
            raise

    @staticmethod
    def id():
        return "demo"

    @staticmethod
    def description():
        return "Demo flow"



from typing import List

from pydash import _


def get_inner_hits(result_set: dict):
    if not result_set:
        result_set = {}
    return (
        result_set.get("inner_hits", {})
        .get("chunks", {})
        .get("hits", {})
        .get("hits", [])
    )


def merge_inner_hits_results(inner_hits: List[dict], orig_doc: dict):
    original_chunk = orig_doc.get("_source", {})
    original_system_id = original_chunk.get("system_id", "")

    # Sort by system_id ascending
    inner_hits.append(orig_doc)
    inner_hits = sorted(_.uniq(inner_hits), key=lambda x: x["_id"])
    inner_hits = _.uniq_by(
        inner_hits, lambda x: x["_source"].get("system_id", x["_id"])
    )
    # text
    merged_text = "\n".join([x["_source"]["text"] for x in inner_hits])

    # base doc
    merged_inner_hits = inner_hits[0]["_source"] if inner_hits else {}

    # metadata
    merged_inner_hits["_score"] = orig_doc.get("_score", 0)
    merged_inner_hits["_index"] = orig_doc.get("_index", "")

    merged_inner_hits["_id"] = original_chunk["id"]
    merged_inner_hits["text"] = merged_text
    merged_inner_hits["chunk_text_length"] = len(merged_text)
    start_system_id = inner_hits[0]["_source"].get("system_id", "")
    end_system_id = inner_hits[-1]["_source"].get("system_id", "")
    merged_inner_hits["system_id"] = (
        f"{start_system_id} - {end_system_id} ({original_system_id})"
    )
    return merged_inner_hits



import os

from fastapi import Depends, FastAPI, HTTPException
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from shraga_common.logging import get_git_commit

from ..auth.backends import (BasicAuthBackend, GoogleAuthBackend,
                             JWTAuthBackend, MicrosoftAuthBackend)
from ..config import get_config
from ..middlewares import logging_middleware
from ..services.analytics_service import is_analytics_authorized

from .analytics_api import router as analytics_router
from .flows_api import router as flows_router
from .history_api import router as history_router
from .services_api import router as services_router

api_app = FastAPI(root_path="/api")

api_app.add_middleware(BaseHTTPMiddleware, dispatch=logging_middleware)

def load_api_app():
    shraga_config = get_config()
    if shraga_config.auth_realms():

        @api_app.middleware("http")
        async def enforce_auth(request: Request, call_next):
            if not request.user.is_authenticated:
                return JSONResponse({"detail": "Unauthenticated"}, status_code=401)
            request.state.user_display_name = request.user.display_name
            response = await call_next(request)
            return response

        if "jwt" in shraga_config.auth_realms():
            api_app.add_middleware(
                AuthenticationMiddleware,
                backend=JWTAuthBackend(),
                on_error=lambda conn, exc: JSONResponse(
                    {"detail": str(exc)}, status_code=401
                ),
            )
        if "basic" in shraga_config.auth_realms():
            api_app.add_middleware(
                AuthenticationMiddleware,
                backend=BasicAuthBackend(),
                on_error=lambda conn, exc: JSONResponse(
                    {"detail": str(exc)}, status_code=401
                ),
            )
        if "google" in shraga_config.auth_realms():
            api_app.add_middleware(
                AuthenticationMiddleware,
                backend=GoogleAuthBackend(),
                on_error=lambda conn, exc: JSONResponse(
                    {"detail": str(exc)}, status_code=401
                ),
            )
        if "microsoft" in shraga_config.auth_realms():
            api_app.add_middleware(
                AuthenticationMiddleware,
                backend=MicrosoftAuthBackend(),
                on_error=lambda conn, exc: JSONResponse(
                    {"detail": str(exc)}, status_code=401
                ),
            )

        @api_app.get("/whoami")
        async def whoami(request: Request) -> dict:
            display_name = (
                request.user.display_name if hasattr(request, "user") else "<unknown>"
            )
            ret = {
                "display_name": display_name,
                "shraga_version": get_git_commit() or "unknown",
                # "identity": request.user.identity,
            }
            if is_analytics_authorized(display_name):
                ret["roles"] = ["analytics"]
            return ret

    else:

        @api_app.get("/whoami")
        async def whoami() -> dict:
            return {"user": "<unknown>", "roles": ["analytics"]}


    if os.getenv("SHRAGA_PROD") != "true":
        # Only enable CORS on non-prod
        api_app.add_middleware(
            CORSMiddleware,
            allow_origins=[
                "http://localhost:5000",
            ],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    api_app.include_router(
        flows_router,
        prefix="/flows",
        tags=["flows"],
    )

    if not get_config("history.enabled") is False:
        api_app.include_router(
            history_router,
            prefix="/history",
            tags=["history"],
        )
        if get_config("history.analytics"):

            def check_analytics_auth(request: Request):
                email = request.user.display_name if hasattr(request, "user") else None
                if not is_analytics_authorized(email):
                    raise HTTPException(status_code=403)
                return True

            api_app.include_router(
                analytics_router,
                prefix="/analytics",
                tags=["analytics"],
                dependencies=[Depends(check_analytics_auth)],
            )


    api_app.include_router(
        services_router,
        prefix="/services",
        tags=["services"],
    )


    @api_app.get("/ui/configs")
    async def ui_configs() -> dict:
        d = get_config("ui")
        d["map_api"] = get_config("services.googlemaps")
        d["input_max_length"] = get_config("flows.input_max_length")
        d["build"] = os.getenv("SHRAGA_BUILD_TAG")
        d["prod"] = os.getenv("SHRAGA_PROD")

        return d



import datetime
import logging
from typing import List, Optional

from fastapi import Request
from opensearchpy import NotFoundError
from pydash import _

from shraga_common.logging import (get_config_info, get_platform_info,
                                   get_user_agent_info)
from shraga_common.models import FlowResponse, FlowStats

from ..utils import is_prod_env
from ..config import get_config
from ..models import Chat, ChatMessage, FeedbackRequest, FlowRunApiRequest
from .get_history_client import get_history_client

logger = logging.getLogger(__name__)


async def get_history(
    user_id: Optional[str], 
    start: Optional[str] = None, 
    end: Optional[str] = None
) -> List[Chat]:
    try:
        shraga_config = get_config()
        client, index = get_history_client(shraga_config)
        if not client:
            return []

        filters = []
        chat_list_length = 50

        if user_id:
            filters.append({"term": {"user_id": user_id}})
            chat_list_length = 10
        if start:
            filters.append(
                {"range": {"timestamp": {"gte": start, "lte": end or "now"}}}
            )

        if is_prod_env() and not user_id:
            filters.append({"term": {"config.prod": True}})

        bool_query = {
            "must": [{"terms": {"msg_type": ["system", "user"]}}],
            "filter": filters,
        }

        query = {
            "query": {
                "bool": bool_query
            },
            "size": 0,
            "aggs": {
                "by_chat": {
                    "terms": {
                        "field": "chat_id",
                        "size": chat_list_length,
                        "order": {"first_message": "desc"},
                    },
                    "aggs": {
                        "latest": {
                            "top_hits": {
                                "size": 100,
                                "sort": [{"timestamp": {"order": "desc"}}],
                            }
                        },
                        "first_message": {"min": {"field": "timestamp"}},
                    },
                }
            },
        }

        response = client.search(
            index=index,
            body=query,
        )
        hits = _.get(response, "aggregations.by_chat.buckets") or []
        return [Chat.from_hit(hit) for hit in hits]
    except NotFoundError:
        logger.error("Error retrieving history (index not found)")
        return []
    except Exception as e:
        logger.exception("Error retrieving history", exc_info=e)
        return []


async def get_chat(chat_id: str) -> Optional[Chat]:
    shraga_config = get_config()
    client, index = get_history_client(shraga_config)
    if not client:
        return None

    try:
        response = client.get(index=index, id=chat_id)
        if not response["found"]:
            return None
        return Chat(**response["_source"])
    except Exception:
        logger.exception("Error retrieving chat")
        return None


async def delete_chat(chat_id: str) -> bool:
    shraga_config = get_config()
    client, index = get_history_client(shraga_config)
    if not client:
        return True

    try:
        client.delete(index=index, id=chat_id)
        return True
    except Exception:
        logger.exception("Error deleting chat")
    return False


async def log_interaction(msg_type: str, request: Request, context: dict):
    try:
        shraga_config = get_config()
        client, index = get_history_client(shraga_config)
        if not client:
            return

        user_id = request.user.display_name if hasattr(request, "user") else "<unknown>"
        message = ChatMessage(
            msg_type=msg_type,
            timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
            user_id=user_id,
            **context
        )

        o = message.model_dump()
        o["platform"] = get_platform_info()
        o["config"] = get_config_info(shraga_config)
        o["user_agent"] = get_user_agent_info(request.headers.get("user-agent"))

        client.index(index=index, body=o)
        return True

    except Exception as e:
        logger.exception("Error logging interation %s", msg_type, exc_info=e)
        return False


async def log_feedback(request: Request, request_body: FeedbackRequest) -> bool:
    return await log_interaction(
        "feedback",
        request,
        {
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": request_body.feedback_text,
            "position": request_body.position,
            "feedback": request_body.feedback,
        },
    )


async def log_user_message(request: Request, request_body: FlowRunApiRequest):
    return await log_interaction(
        "user",
        request,
        {
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": request_body.question,
            "position": request_body.position,
        },
    )


async def log_system_message(
    request: Request,
    request_body: FlowRunApiRequest,
    response: Optional[FlowResponse] = None,
):
    # delete extra before storing history
    if response.retrieval_results:
        for result in response.retrieval_results:
            if result.extra:
                result.extra = {}
    return await log_interaction(
        "system",
        request,
        {
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": response.response_text,
            "position": request_body.position + 1,
            "stats": response.stats,
            "payload": response.payload,
            "retrieval_results": response.retrieval_results,
            "trace": response.trace,
        },
    )


async def log_flow(request: Request, request_body: FlowRunApiRequest, stat: FlowStats):
    return await log_interaction(
        "flow_stats",
        request,
        {
            "chat_id": request_body.chat_id,
            "text": request_body.question,
            "flow_id": stat.flow_id,
            "stats": stat,
        },
    )


async def log_error_message(
    request: Request, request_body: FlowRunApiRequest, error: Exception, traceback: str
):
    return await log_interaction(
        "error",
        request,
        {
            "chat_id": request_body.chat_id,
            "flow_id": request_body.flow_id,
            "text": str(error),
            "traceback": traceback,
        },
    )



from datetime import datetime
from typing import List, Literal, Optional

from pydantic import BaseModel
from pydash import _

from shraga_common.models import FlowStats, RetrievalResult


class ChatMessage(BaseModel):
    timestamp: datetime
    chat_id: str
    flow_id: str
    text: Optional[str] = None
    user_id: Optional[str] = None
    msg_type: Literal["user", "system", "feedback", "flow_stats", "error"]
    position: Optional[int] = None
    context: Optional[dict] = None
    feedback: Optional[str] = None
    stats: Optional[List[FlowStats] | FlowStats] = None
    retrieval_results: Optional[List[RetrievalResult]] = None
    payload: Optional[dict] = None
    trace: Optional[dict] = None
    traceback: Optional[str] = None


class Chat(BaseModel):
    chat_id: str
    timestamp: datetime
    messages: List[ChatMessage] = []
    user_id: Optional[str] = None
    flow_id: Optional[str] = None
    step_stats: Optional[List[FlowStats]] = None
    total_stats: Optional[FlowStats] = None

    @staticmethod
    def from_hit(hit: dict):
        messages = sorted(
            _.get(hit, "latest.hits.hits", []), key=lambda x: x["_source"]["timestamp"]
        )

        latest_message = messages[0]["_source"] if messages else {}
        chat = Chat(
            chat_id=hit.get("key"),
            timestamp=latest_message.get("timestamp"),
            messages=[ChatMessage(**x["_source"]) for x in messages],
            user_id=latest_message.get("user_id"),
            flow_id=latest_message.get("flow_id"),
        )
        return chat



import asyncio
import json
import time
from typing import List, Literal, Optional, TypedDict

import boto3
import botocore

from shraga_common import ShragaConfig

from ..app.exceptions import LLMServiceUnavailableException
from ..models import FlowStats
from ..utils import safe_to_int
from .common import LLMModelResponse
from .llm_service import LLMService, LLMServiceOptions


class BedrockChatModelId(TypedDict):
    cohere_command: str
    haiku: str
    haiku_3_5: str
    sonnet_3: str
    sonnet_3_5: str
    sonnet_3_5_v2: str


BedrockModelNames = Literal[tuple(BedrockChatModelId.__annotations__.keys())]

BEDROCK_CHAT_MODEL_IDS: BedrockChatModelId = {
    # "cohere_command": "cohere.command-text-v14",
    "haiku": "anthropic.claude-3-haiku-20240307-v1:0",
    "sonnet_3": "anthropic.claude-3-sonnet-20240229-v1:0",
    "sonnet_3_5": "anthropic.claude-3-5-sonnet-20240620-v1:0",
    # cross region routing
    "haiku_3_5": "us.anthropic.claude-3-5-haiku-20241022-v1:0",
    "sonnet_3_5_v2": "us.anthropic.claude-3-5-sonnet-20241022-v2:0",
    "sonnet_3_7": "us.anthropic.claude-3-7-sonnet-20250219-v1:0",
    "nova_pro": "amazon.nova-pro-v1:0",
}


class BedrockService(LLMService):
    def __init__(
        self,
        shraga_config: Optional[ShragaConfig] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        profile_name: Optional[str] = None,
        region_name: str = "us-east-1",
    ):

        if shraga_config:
            aws_access_key_id = aws_access_key_id or shraga_config.get(
                "aws.access_key_id"
            )
            aws_secret_access_key = aws_secret_access_key or shraga_config.get(
                "aws.secret_access_key"
            )
            profile_name = profile_name or shraga_config.get("aws.profile")
        if aws_access_key_id == "":
            aws_access_key_id = None
        if aws_secret_access_key == "":
            aws_secret_access_key = None
        self.boto = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=aws_access_key_id if aws_access_key_id else None,
            aws_secret_access_key=(
                aws_secret_access_key if aws_secret_access_key else None
            ),
            # profile=profile_name,
            region_name=region_name,
        )

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "instance"):
            cls.instance = super(BedrockService, cls).__new__(cls)
        return cls.instance

    async def invoke_model(
        self, prompt: str, options: Optional[LLMServiceOptions] = None
    ):
        return await self.invoke_chat_model(prompt, options)

    async def invoke_converse_model(
        self, system_prompt: List[str], prompt: str, tool_config: Optional[dict] = {}, options: dict = {}
    ):
        if not options:
            options = {
                "model_id": "sonnet_3",
            }

        model_name: BedrockModelNames = options.get("model_id")
        model_id = BEDROCK_CHAT_MODEL_IDS[model_name]

        system_messages = [{"text": msg} for msg in system_prompt]
        messages = [{"role": "user", "content": [{"text": prompt}]}]

        converse_params = {
                "modelId": model_id,
                "system": system_messages,
                "messages": messages,
                "inferenceConfig": {"temperature": 0.0, "maxTokens": 8192},
            }
        
        if tool_config:
            converse_params["toolConfig"] = tool_config
        
        try:
            start_time = time.time()
            response = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: self.boto.converse(
                    **converse_params,
                ),
            )
            
            if response.get('Error'):
                error_code = response['Error'].get('Code')
                error_message = response['Error'].get('Message')
                raise LLMServiceUnavailableException(f"Bedrock error ({error_code}, {error_message})")

            stats = FlowStats(
                llm_model_id=model_id,
                time_took=time.time() - start_time,
                latency=response['metrics']['latencyMs'],
                input_tokens=response['usage']['inputTokens'],
                output_tokens=response['usage']['outputTokens'],
                total_tokens=response['usage']['totalTokens'],
            )

            text = response["output"]["message"]["content"][0]["text"]

            return LLMModelResponse(text=text, stats=stats)
        
        except LLMServiceUnavailableException:
            raise
        except (
            boto3.exceptions.Boto3Error,
            boto3.exceptions.RetriesExceededError,
            botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError,
            asyncio.TimeoutError,
        ) as e:
            raise LLMServiceUnavailableException("Bedrock error", e)
        except Exception as e:
            raise Exception(f"Error invoking Bedrock model: {str(e)}")
        

    async def invoke_chat_model(
        self, prompt: str, options: Optional[LLMServiceOptions] = None
    ) -> LLMModelResponse:
        if not options:
            options = {
                "model_id": "sonnet_3",
            }

        model_name: BedrockModelNames = options.get("model_id")
        model_id = BEDROCK_CHAT_MODEL_IDS[model_name]

        body = json.dumps(
            {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": 9182,
                "temperature": 0.0,
                "messages": [
                    {
                        "role": "user",
                        "content": [{"type": "text", "text": prompt}],
                    }
                ],
            }
        )

        try:
            start_time = time.time()
            response = await asyncio.get_event_loop().run_in_executor(
                None,
                lambda: self.boto.invoke_model(
                    body=body,
                    modelId=model_id,
                    accept="application/json",
                    contentType="application/json",
                ),
            )
            time_took = time.time() - start_time
            headers = response.get("ResponseMetadata").get("HTTPHeaders")
            stats = FlowStats(
                llm_model_id=model_id,
                time_took=time_took,
                latency=safe_to_int(headers.get("x-amzn-bedrock-invocation-latency")),
                input_tokens=safe_to_int(headers.get("x-amzn-bedrock-input-token-count")),
                output_tokens=safe_to_int(headers.get("x-amzn-bedrock-output-token-count")),
            )
            body = response.get("body").read()
            text = json.loads(body).get("content")[0].get("text")
            return LLMModelResponse(text=text, stats=stats)
        
        except (
            boto3.exceptions.Boto3Error,
            boto3.exceptions.RetriesExceededError,
            botocore.exceptions.BotoCoreError,
            botocore.exceptions.ClientError,
            asyncio.TimeoutError,
        ) as e:
            raise LLMServiceUnavailableException("Bedrock error", e)
        except Exception as e:
            raise Exception(f"Error invoking Bedrock model: {str(e)}")




import time
from typing import Literal, Optional, TypedDict

from openai import OpenAI, OpenAIError

from shraga_common import ShragaConfig

from ..models import FlowStats
from .common import LLMModelResponse
from .llm_service import LLMService, LLMServiceOptions


class OpenAiModelId(TypedDict):
    gpt_4o: str
    gpt_4o_mini: str
    gpt_4_turbo: str
    turbo_3_5: str


OpenAiModelNames = Literal[tuple(OpenAiModelId.__annotations__.keys())]

OPENAI_CHAT_MODEL_IDS: OpenAiModelId = {
    "gpt_4o": "gpt-4o",  # gpt-4o-2024-05-13
    "gpt_4o_mini": "gpt-4o-mini",  # gpt-4o-mini-2024-07-18
    "gpt_4_turbo": "gpt-4-turbo",  # gpt-4-turbo-2024-04-09
    "turbo_3_5": "gpt-3.5-turbo",  # gpt-3.5-turbo-0125
}


class OpenAIService(LLMService):
    def __init__(
        self,
        shraga_config: Optional[ShragaConfig] = None,
        api_key: Optional[str] = None,
    ):
        if shraga_config:
            api_key = api_key or shraga_config.get("services.openai.api_key")

        self.client = OpenAI(api_key=api_key)

    async def invoke_model(self, prompt: str, options: LLMServiceOptions = None) -> str:
        if not options:
            options = {
                "model_id": "gpt_4o_mini",
            }

        model_name: OpenAiModelId = options.get("model_id")
        model_id = OPENAI_CHAT_MODEL_IDS[model_name]
        system_prompt = options.get("system_prompt", None)

        try:
            messages = []
            if system_prompt:
                messages.append({"role": "system", "content": system_prompt})
            messages.append({"role": "user", "content": prompt})

            start_time = time.time()
            response = self.client.chat.completions.create(
                model=model_id, messages=messages
            )
            time_took = time.time() - start_time
            text = response.choices[0].message.content

            return LLMModelResponse(
                text=text,
                stats=FlowStats(
                    llm_model_id=model_id,
                    time_took=time_took,
                    input_tokens=len(prompt),
                    output_tokens=len(text),
                ),
            )

        except OpenAIError as e:
            # Handle OpenAI API errors
            raise RuntimeError(f"OpenAI API returned an error: {e}")

        except Exception as e:
            # Handle other unforeseen errors
            raise RuntimeError(f"An unexpected error occurred: {e}")


# Usage:
# service = OpenAIService(api_key="your_api_key_here")
# response = service.invoke_model(prompt="Your prompt here")
# print(response)



from typing import Optional, TypedDict


class RetrieverConfig(TypedDict):
    type: str
    auth_method: str
    host: str
    index: str
    port: Optional[int]
    user: Optional[str]
    password: Optional[str]
    use_ssl: Optional[bool]
    verify_certs: Optional[bool]
    use_cloud_id: Optional[bool]



import logging
from typing import List, Optional

from fastapi import APIRouter, BackgroundTasks, HTTPException, Request

from shraga_common.models import FlowResponse

from ..config import get_config
from ..models import FlowRunApiRequest
from ..services import list_flows_service
from .run_flow import run_flow

router = APIRouter()

logger = logging.getLogger(__name__)


@router.post("/run")
async def run_flow_api(
    request: Request,
    req_body: FlowRunApiRequest,
    bg_tasks: BackgroundTasks,
    keep: bool = True
) -> FlowResponse:
    return await run_flow(request, req_body, bg_tasks, keep)

@router.get("/")
async def get_flows_list(full: Optional[bool] = True) -> List[dict]:
    shraga_config = get_config()
    routes = []
    list_flows = get_config("ui.list_flows")
    default_flow = get_config("ui.default_flow")
    
    flows = list_flows_service.get_flows(shraga_config)

    if list_flows:
        for k, v in flows.items():
            if v["obj"].listed:
                routes.append(
                    {
                        "id": k,
                        "description": v.get("description"),
                        "preferences": v.get("preferences") if full else None,
                    }
                )
        routes.sort(key=lambda x: x["id"])
    else:
        if not default_flow:
            return routes

        default_flows = (
            [default_flow] if isinstance(default_flow, str) else default_flow
        )

        for flow_id in default_flows:
            if flow_id in flows:
                routes.append(
                    {
                        "id": flow_id,
                        "description": flows[flow_id].get("description"),
                        "preferences": (
                            flows[flow_id].get("preferences") if full else None
                        ),
                    }
                )
        routes.sort(key=lambda x: x["id"])

    return routes


@router.get("/{flow_id}/preferences")
async def get_flow_preferences(flow_id: str) -> dict:
    if get_config("ui.list_flows") is not True:
        raise HTTPException(status_code=404)

    f = list_flows_service.get_flow(flow_id)
    if not f:
        raise HTTPException(status_code=400, detail="Unknown flow id: " + flow_id)

    return f.get("preferences") or {}




import json
import logging
import os
from typing import Dict, List, Optional

import snowflake.connector
from sentence_transformers import SentenceTransformer

from shraga_common import ShragaConfig

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SnowflakeHuggingfaceEmbedder:
    def __init__(
        self,
        shraga_config: Optional[ShragaConfig] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        account: Optional[str] = None,
        warehouse: Optional[str] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        model_name_or_path: Optional[str] = "all-MiniLM-L6-v2",
    ):

        # Initialize Snowflake connection
        snowflake_configs: Dict = (
            shraga_config.get("embedders.snowflake_huggingface")
            if shraga_config
            else {}
        )

        if snowflake_configs:
            user = user or snowflake_configs.get("user")
            password = password or snowflake_configs.get("password")
            account = account or snowflake_configs.get("account")
            warehouse = warehouse or snowflake_configs.get("warehouse")
            database = database or snowflake_configs.get("database")
            schema = schema or snowflake_configs.get("schema")

        self.snowflake_client = snowflake.connector.connect(
            user=user,
            password=password,
            account=account,
            warehouse=warehouse,
            database=database,
            schema=schema,
        )

        # Load a SentenceTransformer model (appropriate for sentence embeddings)
        self.model = SentenceTransformer(model_name_or_path)

    def fetch_data(self, query: str) -> List[str]:
        """Fetch text data from Snowflake."""
        try:
            with self.snowflake_client.cursor() as cur:
                cur.execute(query)
                results = cur.fetchall()
                return [row[0] for row in results]
        except snowflake.connector.Error as e:
            logger.error(f"Snowflake error occurred: {e}")
            raise

    def generate_embeddings(self, texts: List[str]) -> List[List[float]]:
        """Generate embeddings for a list of texts using a Hugging Face model."""
        try:
            embeddings = self.model.encode(texts)
            return embeddings.tolist()
        except Exception as e:
            logger.error(f"Error generating embeddings: {e}")
            raise

    def store_embeddings(
        self, texts: List[str], embeddings: List[List[float]], table: str
    ):
        """Store embeddings back into Snowflake."""
        try:
            with self.snowflake_client.cursor() as cur:
                for text, embedding in zip(texts, embeddings):
                    embedding_json = json.dumps(embedding)
                    cur.execute(
                        f"INSERT INTO {table} (text_column, embedding_column) VALUES (%s, %s::VARIANT)",
                        (text, embedding_json),
                    )
                self.snowflake_client.commit()
        except snowflake.connector.Error as e:
            logger.error(f"Error storing embeddings in Snowflake: {e}")
            raise

    def process_and_store(self, fetch_query: str, dest_table: str):
        """Fetch data from Snowflake, generate embeddings, and store them back."""
        try:
            # Fetch data
            data_list = self.fetch_data(fetch_query)
            logger.info(f"Fetched {len(data_list)} records from Snowflake.")

            # Generate embeddings
            embeddings_list = self.generate_embeddings(data_list)

            # Store embeddings back into Snowflake
            self.store_embeddings(data_list, embeddings_list, dest_table)
            logger.info(
                f"Stored {len(embeddings_list)} embeddings back into Snowflake."
            )
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            raise


# Example usage
# if __name__ == "__main__":
#    embedder = SnowflakeHuggingfaceEmbedder()
#
#    try:
#        fetch_query = "SELECT text_column FROM your_source_table"
#        dest_table = "your_table_with_embeddings"
#        embedder.process_and_store(fetch_query, dest_table)
#    except Exception as e:
#        logger.error(f"Error in processing and storing embeddings: {e}")



from typing import Literal, Optional

from pydantic import BaseModel


class FeedbackRequest(BaseModel):
    chat_id: str
    flow_id: str
    user_id: str
    position: Optional[int]
    feedback: Literal["thumbs_up", "thumbs_down"]
    feedback_text: Optional[str] = None



from .chunkers import BasicChunker, ChunkOptions
from .cleaners import BasicCleaner, normalize_url
from .html_chunker import HtmlChunker
from .ingest import IngestUtils
from .parsers import BedrockParser, extract_xml_section
from .typing import safe_to_int

__all__ = [
    "BedrockParser",
    "BasicChunker",
    "HtmlChunker",
    "BasicCleaner",
    "ChunkOptions",
    "IngestUtils",
    "safe_to_int",
    "normalize_url",
    "extract_xml_section",
]



from pathlib import Path

from shraga_common.models import FlowBase

PACKAGE_BASE_PATH = Path(__file__).parent.parent.parent

flow_classes = []
flows = {}
available_flows = {}

def register_flows(flws: list[FlowBase], shraga_config):
    global flows, available_flows, flow_classes
    for flow in flws:
        if not issubclass(flow, FlowBase):
            raise TypeError(
                f"Flow class {flow.__name__} must be a subclass of FlowBase"
            )
    flow_classes = flws  
    flows = get_flows(shraga_config)
    available_flows = {k: v.get("obj") for k, v in flows.items() if v and v.get("obj")}
    
def get_flows(shraga_config):
    global flow_classes
    return {
        f.id(): {
            "description": f.description(),
            "preferences": f.available_preferences(shraga_config),
            "obj": f,
        }
        for f in flow_classes
    }
    
def get_flow(flow_id):
    global flows
    flow = flows.get(flow_id)
    if not flow:
        raise RuntimeError(f"No flow found with id {flow_id}")
    return flow

def get_available_flows():
    global available_flows
    return available_flows
    


class BasicCleaner:
    def clean(self, texts) -> list[list]:
        return [text.strip() for text in texts]


def normalize_url(url):
    """
    Normalizing url to the same format.

    Args:
        url (str): the url to normalize.

    Returns:
        Normalized url.
    """
    if url and url[-1] == "/":
        return url[:-1]
    return url



from .config import get_config
from .exceptions import RequestCancelledException, LLMServiceUnavailableException
from .routers import setup_app

__all__ = ['RequestCancelledException', 'LLMServiceUnavailableException', 'setup_app', 'get_config']



from typing import Optional

from pydantic import BaseModel


class RetrievalResult(BaseModel):
    id: Optional[str] = None
    document_id: Optional[int] = None
    date: Optional[str] = None
    title: Optional[str] = "Result"
    link: Optional[str] = None
    description: Optional[str] = None
    score: Optional[float] = None
    extra: Optional[dict] = None



import gzip
import json
import os
import zipfile
from json import JSONDecodeError
from typing import Literal


class LocalRetriever:

    @staticmethod
    def get_data_from_folder(
        path: str,
        verbose: bool = False,
        encoding_method: str = "utf8",
        file_type: Literal["json", "zip"] = "json",
    ):
        """
        Get json data from each file in a folder.

        Args:
            path (str): Folder's full path.
            verbose (bool): Add prints.
            file_type (json \ zip): Type of files to load.

        Returns:
            dict: Parsed json data by relevant keys.
        """

        if not os.path.exists(path) or not os.path.isdir(path):
            raise ValueError(f"Invalid folder path: {path}")

        for file_name in os.listdir(path):
            print("Loading file", file_name)
            if file_name.endswith(f".{file_type}"):
                file_path = os.path.join(path, file_name)

                if file_type == "json":
                    yield from LocalRetriever.get_data_from_json_file(
                        file_path, verbose=verbose, encoding_method=encoding_method
                    )
                elif file_type == "zip":
                    yield from LocalRetriever.get_data_from_zip(
                        file_path, verbose=verbose, encoding_method=encoding_method
                    )

    @staticmethod
    def get_data_from_zip(
        zip_path: str, verbose: bool = False, encoding_method: str = "utf8"
    ):
        archive = zipfile.ZipFile(zip_path, "r")
        for name in archive.namelist():
            if name.endswith(".json") and not name.startswith("__MACOSX"):
                if verbose:
                    print("Loading file", name)
                with archive.open(name) as f:
                    try:
                        content = json.load(f)
                        if isinstance(content, list):
                            for item in content:
                                yield item
                        else:
                            yield content
                    except JSONDecodeError:
                        print(f"Error loading {name} - JSONDecodeError")

    @staticmethod
    def get_data_from_gz(
        gz_path: str, verbose: bool = False, encoding_method: str = "utf8"
    ):
        with gzip.open(gz_path, "rt", encoding=encoding_method) as f:
            if verbose:
                print(f"Loading file {gz_path}")
            try:
                for line in f:
                    yield json.loads(line)
            except JSONDecodeError:
                print(f"Error loading {gz_path} - JSONDecodeError")

    @staticmethod
    def get_data_from_json_file(
        file_path: str, verbose: bool = False, encoding_method: str = "utf8"
    ):
        with open(file_path, "r", encoding=encoding_method) as f:
            if verbose:
                print(f"Loading file {file_path}")
            try:
                content = json.load(f)
                yield content
            except JSONDecodeError:
                print(f"Error loading {file_path} - JSONDecode")

    @staticmethod
    def get_data_from_folders(path: str, verbose: bool = True):
        """
        Get json data from each file within a folder of folders.

        Args:
            path (str): Folder's full path.
            verbose (bool): Add prints.

        Returns:
            dict: Parsed json data by relevant keys.
        """
        # Only folders
        for sub_folder in [
            f_name for f_name in os.listdir(path) if not f_name.startswith(".")
        ]:
            if verbose:
                print(sub_folder)
            yield LocalRetriever.get_data_from_folder("/".join([path, sub_folder]))



import json
from abc import abstractmethod
from datetime import datetime
from typing import Optional

from shraga_common import ShragaConfig
from shraga_common.app import RequestCancelledException, LLMServiceUnavailableException
from shraga_common.models import FlowResponse, FlowRunRequest
from shraga_common.services import LLMServiceOptions

from .llm_flow_base import LLMFlowBase


class AgenticPlanFlowBase(LLMFlowBase):
    llm_model_provider = "cohere"
    llm_model_name = "sonnet_3_5_v2"
    available_tools = []

    def __init__(self, config: ShragaConfig, flows: Optional[dict] = None):
        super().__init__(config, flows)
        self.tool_spec = self.tools_to_spec()

    def tools_to_spec(self):
        parsed_tool_list = []

        for tool in self.available_tools:
            tool_details = tool.get_tool_desc()
            tool_name = tool_details["flow_name"]
            tool_desc = tool_details["description"]
            tool_schema = tool_details["schema"]

            tool_json = {
                "toolSpec": {
                    "name": tool_name,
                    "description": tool_desc,
                    "inputSchema": {"json": tool_schema},
                }
            }

            parsed_tool_list.append(tool_json)

        tool_config = {"tools": parsed_tool_list}
        return tool_config

    @abstractmethod
    def get_prompt(self):
        pass

    @abstractmethod
    def format_prompt(self, question: str, format_info: Optional[dict] = None) -> str:
        pass

    @abstractmethod
    def get_system_prompts(self, request: FlowRunRequest):
        pass

    def get_chat_history(self, request: FlowRunRequest):
        chat_history = request.chat_history
        history_window = request.preferences.get("history_window", 0)
        return LLMFlowBase.format_chat_history(chat_history, history_window)

    async def execute(self, request: FlowRunRequest) -> FlowResponse:
        content = None
        response_text = ""
        payload = {}
        llm_context: LLMServiceOptions = {}
        if self.llm_model_name:
            llm_context["model_id"] = self.llm_model_name

        self.init_model()

        prompt = self.format_prompt(
            request.question,
            {
                **(
                    request.context
                    if isinstance(request.context, dict)
                    else request.context.dict()
                ),
                "chat_history": self.get_chat_history(request),
            },
        )

        system_prompts = self.get_system_prompts(request)
        start_time = datetime.now()

        try:
            content = await self.llmservice.invoke_converse_model(
                system_prompts, prompt, self.tool_spec, llm_context
            )
            run_time = datetime.now() - start_time
            self.trace(f"execute runtime: {run_time}")
            if self.parse_json:
                payload = json.loads(content.text, strict=False)
                payload = payload.get("plan", [])
            else:
                response_text = content.text

        except (RequestCancelledException, LLMServiceUnavailableException):
            raise
        except Exception as e:
            payload = {"error": str(e), "body": content if content else ""}

        return FlowResponse(
            response_text=response_text,
            payload=payload,
            trace=self.trace_log,
            stats=[self.get_stats(content)],
        )




import asyncio
from typing import Dict, List, Optional

import boto3
from opensearchpy import AWSV4SignerAuth, OpenSearch, RequestsHttpConnection
from pydash import _

from shraga_common import ShragaConfig

from .base_search_retriever import BaseSearchRetriever
from .common import RetrieverConfig


class OpenSearchRetriever(BaseSearchRetriever):
    """
    OpenSearchRetriever class
    """

    def __init__(self, shraga_config: ShragaConfig):
        super().__init__()
        self.config = shraga_config
        # Retrieve configurations from environment variables
        config: RetrieverConfig = self.config.get("retrievers.opensearch")
        self.client = OpenSearchRetriever.get_client(shraga_config, config)
        self.index_name = config.get("index")

    @staticmethod
    def get_client(shraga_config, extra_configs: RetrieverConfig):
        if not extra_configs:
            extra_configs = {}
        host = extra_configs.get("host")
        port = extra_configs.get("port")
        use_ssl = True

        verify_certs = True
        if extra_configs.get("use_ssl") is False:
            use_ssl = False
        else:
            verify_certs = (
                extra_configs.get("verify_certs")
                and extra_configs.get("verify_certs") is True
            )

        auth_method = shraga_config.get("aws.auth_method", "basic")
        if auth_method == "aws":
            credentials = boto3.Session().get_credentials()
            region = shraga_config.get("aws.region") or "us-east-1"
            auth = AWSV4SignerAuth(credentials, region, "es")
        else:
            http_auth_user = extra_configs.get("user")
            http_auth_password = extra_configs.get("password")
            auth = (http_auth_user, http_auth_password)

        return OpenSearch(
            hosts=[{"host": host, "port": port}] if port else [host],
            http_auth=auth,
            use_ssl=use_ssl,
            verify_certs=verify_certs,
            connection_class=RequestsHttpConnection,
        )

    async def get_indices_list(self):
        return self.client.cat.indices()

    async def execute_vector_search(
        self,
        field_name: str,
        query_vector: List[float],
        k: int = 10,
        index_name: Optional[str] = None,
    ) -> List[Dict]:
        query = {
            "size": k,
            "_source": {"excludes": ["vector_*", field_name]},
        }

        query_knn = {"knn": {field_name: {"vector": query_vector, "k": k}}}

        query["query"] = query_knn

        return await self.execute_raw_search(query, index_name)

    async def execute_text_search(
        self,
        text: str,
        k: int = 10,
        index_name: Optional[str] = None,
    ) -> List[Dict]:
        query = {"size": k, "query": {"match": {"content": text}}}
        return await self.execute_raw_search(query, index_name)

    def execute_with_timeout(self, body: dict, index_name: str, timeout):
        return self.client.search(body, index_name or self.index_name)

    async def execute_empty_query(self, index_name: Optional[str] = None):
        return await self.execute_raw_search(
            {"query": {"match_all": {}}, "size": 0}, index_name
        )

    async def execute_raw_search(
        self, raw_query: dict, index_name: Optional[str] = None
    ) -> List[Dict]:
        loop = asyncio.get_event_loop()
        response = await loop.run_in_executor(
            None,
            self.execute_with_timeout,
            raw_query,
            index_name or self.index_name,
            300,
        )
        # TODO validate search response
        hits = _.get(response, "hits.hits") or []
        return hits



from typing import List

from fastapi import APIRouter, HTTPException, Request

from ..models import Chat, FeedbackRequest
from ..services import history_service
from ..utils import non_ok_response, ok_response

router = APIRouter()


@router.get("/", response_model=List[Chat])
async def get_history(request: Request) -> List[Chat]:
    if not request.user.display_name:
        return []
    return await history_service.get_history(request.user.display_name)


@router.get("/{chat_id}", response_model=Chat)
async def get_chat(chat_id: str) -> Chat:
    chat = await history_service.get_chat(chat_id)
    if chat is None:
        raise HTTPException(status_code=404, detail="Chat not found")
    return chat


@router.delete("/{chat_id}")
async def delete_chat(chat_id: str):
    success = await history_service.delete_chat(chat_id)
    if success:
        return ok_response()

    return non_ok_response("Failed to delete chat")


@router.post("/feedback")
async def post_feedback(request: Request, feedback: FeedbackRequest):
    # TODO check auth
    success = await history_service.log_feedback(request, feedback)
    if success:
        return ok_response()

    return non_ok_response("Failed to update feedback")



import base64
import binascii

import jwt
import requests
from starlette.authentication import (AuthCredentials, AuthenticationBackend,
                                      AuthenticationError, SimpleUser)

from ..config import get_config



class BasicAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        if "user" in conn and conn.user.is_authenticated:
            return AuthCredentials(["authenticated"]), conn.user

        if "Authorization" not in conn.headers:
            raise AuthenticationError("Unauthenticated")

        auth = conn.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != "basic":
                return
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise AuthenticationError("Invalid basic auth credentials")

        username, _, password = decoded.partition(":")
        username = username.lower().strip()
        shraga_config = get_config()
        if not (
            username in shraga_config.auth_users()
            and f"{username}:{password}" in shraga_config.auth_realms().get("basic")
        ):
            raise AuthenticationError("Authentication failed")

        return AuthCredentials(["authenticated"]), SimpleUser(username)


class JWTAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        shraga_config = get_config()
        if "user" in conn and conn.user.is_authenticated:
            return AuthCredentials(["authenticated"]), conn.user

        if "Authorization" not in conn.headers:
            raise AuthenticationError("Unauthenticated")

        auth = conn.headers["Authorization"]
        try:
            scheme, token = auth.split()
            if scheme.lower() != "bearer":
                return
            auth_secret = shraga_config.auth_realms().get("jwt").get("secret")
            decoded = jwt.decode(token, auth_secret, algorithms=["HS256"])
        except (ValueError, UnicodeDecodeError, binascii.Error, jwt.DecodeError):
            raise AuthenticationError("Invalid JWT token")

        username = decoded.get("username") or decoded.get("email") or "anonymous"
        return AuthCredentials(["authenticated"]), SimpleUser(str(username).strip())


class GoogleAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        if "user" in conn and conn.user.is_authenticated:
            return AuthCredentials(["authenticated"]), conn.user

        if "Authorization" not in conn.headers:
            raise AuthenticationError("Unauthenticated")

        auth = conn.headers["Authorization"]
        try:
            scheme, token = auth.split()
            if scheme.lower() != "google":
                return
            response = requests.get(
                f"https://www.googleapis.com/oauth2/v1/userinfo?access_token={token}"
            )
            if response.status_code != 200:
                raise AuthenticationError("Invalid Google OAuth token")
            user_info = response.json()
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise AuthenticationError("Invalid Google OAuth token")

        return AuthCredentials(["authenticated"]), SimpleUser(user_info["email"])


class MicrosoftAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        if "user" in conn and conn.user.is_authenticated:
            return AuthCredentials(["authenticated"]), conn.user

        if "Authorization" not in conn.headers:
            raise AuthenticationError("Unauthenticated")

        auth = conn.headers["Authorization"]
        try:
            scheme, token = auth.split()
            if scheme.lower() != "microsoft":
                return
            response = requests.get(
                f"https://graph.microsoft.com/v1.0/me",
                headers={"Authorization": f"Bearer {token}"},
            )
            if response.status_code != 200:
                raise AuthenticationError("Invalid Microsoft OAuth token")
            user_info = response.json()
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise AuthenticationError("Invalid Microsoft OAuth token")

        return AuthCredentials(["authenticated"]), SimpleUser(
            user_info["userPrincipalName"]
        )



PART_EVALUATION_PROMPT = """
You are assisting a human in evaluating an AI-generated response to their question. 
You are given the question, the expected answer and the AI assistant's generated answer, 
Your goal is to assess the relevance, clarity, and completeness of the AI's answer, identifying any errors, ambiguities, or areas for improvement. 
Provide constructive feedback to enhance the quality of the response.
Your evaluation should focus on the content's meaning and essence, not minor details like typos, punctuation or slight differences in phrasing.
Based on your assessment, you are required to determine if the AI assistant's answer is correct or not. 
Assume that each question must and can be answered and the context must have the relevant information to the question.

Correctness labels are defined as follows:
- correct: The answer directly answers the question and is similar to the expected answer in terms relevance and coherence. A correct answer does not have to fully match the level of detail and specificity present in the expected answer as long as it answers the question.
- incorrect: The answer does not directly answers the question or is irrelevant to the the expected answer in terms of topic, relevance and coherence.

```json
{{
    "thought": "explain here step by step why you think the answer is correct or incorrect.",
    "correctness": "correct/incorrect", 
}}

Question: {question}
Expected answer: {expected_answer}
Generated answer: {answer}

"""



from typing import Generic, List, Literal, Optional, TypeVar

from pydantic import BaseModel, Field

# model or default of dict
T = TypeVar("T")


class HistoryMessage(BaseModel):
    text: str
    msg_type: Literal["system", "user"]
    timestamp: Optional[str] = None


class FlowRunRequest(BaseModel, Generic[T]):
    question: Optional[str] = None
    context: T = Field(default_factory=dict)
    chat_history: Optional[List[HistoryMessage]] = []
    preferences: Optional[dict] = {}



import argparse
import os


def load_args():
    parser = argparse.ArgumentParser(description="Run the Ingest Flow")
    parser.add_argument(
        "--input_dir",
        required=True,
        type=str,
        help="Input dir",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=10000,
        help="Batch size",
    )
    parser.add_argument(
        "--target_index",
        type=str,
        help="Target ES index override",
    )
    parser.add_argument(
        "--action",
        type=str,
        help="index or update",
    )
    parser.add_argument(
        "--dry_run",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Dry run",
    )
    parser.add_argument(
        "--doc_type",
        type=str,
        required=True,
        help="Document type",
    )
    parser.add_argument(
        "--max_items",
        type=int,
        default=0,
        help="Max items to ingest",
    )
    args = parser.parse_args()

    if not os.path.isdir(args.input_dir):
        raise ValueError(f"Input dir {args.input_dir} does not exist")

    print(args)
    return args



from typing import Generic, List, Optional, TypeVar
from uuid import uuid4

from pydantic import BaseModel, Field
from pydash import _

from .flow_stats import FlowStats
from .retrieval_result import RetrievalResult

T = TypeVar("T")


class FlowResponse(BaseModel, Generic[T]):
    chat_id: str = Field(default_factory=lambda: uuid4().hex)
    response_text: Optional[str] = None
    allow_reply: bool = False
    retrieval_results: Optional[List[RetrievalResult]] = None
    payload: T = Field(default_factory=dict)
    trace: Optional[dict] = None
    stats: Optional[List[FlowStats]] = None

    def get(self, path: str, default=None):
        if self.payload:
            return _.get(self.payload, path, default)
        return None



from typing import List, Optional

from fastapi import APIRouter, Query

from ..models import AnalyticsRequest, Chat
from ..services import analytics_service, history_service

router = APIRouter()


@router.post("/")
async def get_analytics(r: AnalyticsRequest) -> dict:
    return await analytics_service.get_analytics(r)


@router.get("/chat-history", response_model=List[Chat])
async def get_chat_history(
    start: Optional[str] = Query(None), end: Optional[str] = Query(None)
) -> List[Chat]:
    return await history_service.get_history(None, start, end)



import asyncio
import logging
from typing import Optional

from fastapi import Request

from shraga_common.models import FlowBase, FlowResponse

from ..exceptions import RequestCancelledException, LLMServiceUnavailableException
from ..models import FlowRunApiRequest

logger = logging.getLogger(__name__)


class RequestCancellationManager:
    def __init__(self, request: Request):
        self.request = request
        self.cancel_event = asyncio.Event()
        self._monitor_task: Optional[asyncio.Task] = None

    async def setup(self) -> asyncio.Event:
        async def _monitor():
            try:
                message = await self.request.receive()
                if message["type"] == "http.disconnect":
                    self.cancel_event.set()
            except Exception as e:
                self.cancel_event.set()

        self._monitor_task = asyncio.create_task(_monitor())
        return self.cancel_event

    async def cleanup(self):
        if self._monitor_task and not self._monitor_task.done():
            self._monitor_task.cancel()


async def execute_cancellable_flow(
    request: Request, flow: FlowBase, req_body: FlowRunApiRequest
) -> FlowResponse:
    cancel_manager = RequestCancellationManager(request)

    try:
        cancel_event = await cancel_manager.setup()
        flow_task = asyncio.create_task(flow.execute(req_body))
        cancel_task = asyncio.create_task(cancel_event.wait())

        try:
            done, pending = await asyncio.wait(
                [flow_task, cancel_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if flow_task not in done:
                flow_task.cancel()
                raise RequestCancelledException("Request cancelled by client")

            try:
                return await flow_task
            except LLMServiceUnavailableException:
                raise

        finally:
            for task in [flow_task, cancel_task]:
                if not task.done():
                    task.cancel()
                    try:
                        await task
                    except (
                        asyncio.CancelledError, 
                        RequestCancelledException, 
                        LLMServiceUnavailableException
                    ):
                        pass

    finally:
        await cancel_manager.cleanup()
